from mpopt.utils import if_debug
from mpopt.common import gurobi
from mpopt.mwis.model import Model

import json
import math
import sys


class Callback:
    def __init__(self, time_threshold=10, min_time_diff=1):
        self.time_threshold = time_threshold
        self.min_time_diff = min_time_diff
        self.print_time = None
        self.buffer = None

    def _check_invariants(self, current_time=None):
        if (self.print_time is not None) and (self.buffer is not None):
            buffer_time, _, _ = self.buffer
            assert self.print_time <= buffer_time
        if (self.buffer is not None) and (current_time is not None):
            buffer_time, _, _ = self.buffer
            assert buffer_time <= current_time

    def _update_buffer(self, time, ub, lb):
        if self.buffer is not None:
            _, buffer_ub, buffer_lb = self.buffer
            ub = min(buffer_ub, ub)
            lb = max(buffer_lb, lb)
        self.buffer = (time, ub, lb)

    def _dump(self, opt=False):
        if self.buffer is not None:
            time, ub, lb = self.buffer
            obj = {'t': time, 'ub': ub, 'lb': lb, 'opt': opt}
            json.dump(obj, sys.stdout)
            print(flush=True)
            self.buffer = None
            self.print_time = time

    def __call__(self, model, where):
        time, ub, lb = None, math.inf, -math.inf
        if where == gurobi.GRB.Callback.SIMPLEX:
            time = model.cbGet(gurobi.GRB.Callback.RUNTIME)
            if model.Params.Method == 0:
                lb = model.cbGet(gurobi.GRB.Callback.SPX_OBJVAL)
            else:
                ub = model.cbGet(gurobi.GRB.Callback.SPX_OBJVAL)
        elif where == gurobi.GRB.Callback.BARRIER:
            time = model.cbGet(gurobi.GRB.Callback.RUNTIME)
            lb = model.cbGet(gurobi.GRB.Callback.BARRIER_PRIMOBJ)
            ub = model.cbGet(gurobi.GRB.Callback.BARRIER_DUALOBJ)
        elif where == gurobi.GRB.Callback.MIP:
            time = model.cbGet(gurobi.GRB.Callback.RUNTIME)
            lb = model.cbGet(gurobi.GRB.Callback.MIP_OBJBST)
            ub = model.cbGet(gurobi.GRB.Callback.MIP_OBJBND)
        elif where == gurobi.GRB.Callback.MIPSOL:
            time = model.cbGet(gurobi.GRB.Callback.RUNTIME)
            lb = model.cbGet(gurobi.GRB.Callback.MIPSOL_OBJBST)
            ub = model.cbGet(gurobi.GRB.Callback.MIPSOL_OBJBND)
        elif where == gurobi.GRB.Callback.MIPNODE:
            time = model.cbGet(gurobi.GRB.Callback.RUNTIME)
            lb = model.cbGet(gurobi.GRB.Callback.MIPNODE_OBJBST)
            ub = model.cbGet(gurobi.GRB.Callback.MIPNODE_OBJBND)

        if time is not None:
            self._check_invariants(time)
            if time < self.time_threshold or self.print_time is None:
                self._update_buffer(time=time, ub=ub, lb=lb)
                self._dump(opt=False)
            elif time - self.print_time < self.min_time_diff:
                self._update_buffer(time=time, ub=ub, lb=lb)
            else:
                self._dump(opt=False)
                self._update_buffer(time=time, ub=ub, lb=lb)

    def finalize(self, time, lb, ilp_mode, opt):
        self._dump()
        ub = lb if opt else math.inf
        self._update_buffer(time, ub, lb)
        self._dump(opt=opt)

    def make_function_wrapper(self):
        def callback(model, where):
            return self(model, where)
        return callback


def parse_gurobi_method(text):
    if text == 'primal-simplex':
        return 0
    elif text == 'dual-simplex':
        return 1
    elif text == 'barrier':
        return 2
    else:
        raise RuntimeError(f'Unknown Gurobi method: {text}')


class Gurobi:

    def __init__(self, model, ilp_mode=True, pairwise_relaxation=False, silent=True, method='barrier', timeout=None):
        self._constructed = False
        self.model = model
        self.ilp_mode = ilp_mode
        self.pairwise_relaxation = pairwise_relaxation
        self.gurobi = gurobi.Model()

        if silent:
            self.gurobi.Params.OutputFlag = 0

        self.gurobi.Params.Method = parse_gurobi_method(method)
        self.gurobi.Params.DisplayInterval = 1
        self.gurobi.Params.MIPGap = 0
        self.gurobi.Params.MIPGapAbs = 0
        self.gurobi.Params.Threads = 1

        if timeout is not None:
            self.gurobi.Params.TimeLimit = timeout

    def construct(self):
        if self._constructed:
            return

        vtype = gurobi.GRB.CONTINUOUS
        if self.ilp_mode:
            vtype = gurobi.GRB.BINARY

        self.node_vars = self.gurobi.addVars(
            len(self.model.nodes),
            lb=0.0,
            ub=1.0,
            vtype=vtype,
            obj=self.model.nodes)

        if not self.pairwise_relaxation:
            self.clique_vars = self.gurobi.addVars(
                len(self.model.cliques), lb=0.0, ub=1.0, vtype=vtype, obj=0.0)

            for cidx, clique in enumerate(self.model.cliques):
                lhs = sum(self.node_vars[nidx]
                          for nidx in clique) + self.clique_vars[cidx]
                self.gurobi.addConstr(lhs == 1)
        else:
            number_of_edges = sum(
                (len(clique) * (len(clique) - 1)) // 2
                for clique in self.model.cliques)

            self.edge_vars = self.gurobi.addVars(
                number_of_edges, lb=0.0, ub=1.0, vtype=vtype, obj=0.0)

            edge_index = 0
            for cidx, clique in enumerate(self.model.cliques):
                for i in range(0, len(clique)):
                    for j in range(i + 1, len(clique)):
                        lhs = sum([
                            self.node_vars[clique[i]],
                            self.node_vars[clique[j]],
                            self.edge_vars[edge_index],
                        ])
                        self.gurobi.addConstr(lhs == 1)
                        edge_index += 1
            assert edge_index == number_of_edges

        self.gurobi.ObjCon = self.model.constant
        self.gurobi.ModelSense = gurobi.GRB.MAXIMIZE
        self._constructed = True

    def use_mapping(self, mapping):
        for v, x in zip(self.node_vars.values(), mapping):
            if x in (0, 1):
                self.gurobi.addConstr(v == x)

    def solve(self):
        self.construct()
        cb = Callback()
        self.gurobi.optimize(cb.make_function_wrapper())
        cb.finalize(time=self.gurobi.Runtime, lb=self.gurobi.ObjVal,
                    ilp_mode=self.ilp_mode, opt=self.gurobi.Status == gurobi.GRB.OPTIMAL)
        self.assignment = [var.X for var in self.node_vars.values()]
        if self.ilp_mode:
            self.assignment = [int(x > .5) for x in self.assignment]

    def write(self, filename):
        self.construct()
        self.gurobi.write(filename)
