#ifndef LIBMPOPT_COMMON_MISC_HPP
#define LIBMPOPT_COMMON_MISC_HPP

namespace mpopt {

template<typename T>
struct comparator_min {
  bool operator()(const T& a, const T& b) const {
    return a < b;
  }

  T neutral() const {
    static_assert(std::numeric_limits<T>::has_infinity);
    return std::numeric_limits<T>::infinity();
  }
};

template<typename T>
struct comparator_max {
  bool operator()(const T& a, const T& b) const {
    return a > b;
  }

  T neutral() const {
    static_assert(std::numeric_limits<T>::has_infinity);
    return -std::numeric_limits<T>::infinity();
  }
};

template<typename FORWARD_ITERATOR_DATA, typename FORWARD_ITERATOR_BOOL, typename COMPARATOR>
auto best_element(FORWARD_ITERATOR_DATA data_begin, FORWARD_ITERATOR_DATA data_end,
                  FORWARD_ITERATOR_BOOL active_begin, FORWARD_ITERATOR_BOOL active_end,
                  COMPARATOR comparator)
{
  assert(std::distance(active_begin, active_end) >= std::distance(data_begin, data_end));
  auto minimum = data_end;

  auto data_it = data_begin;
  auto active_it = active_begin;
  for (; data_it != data_end; ++data_it, ++active_it) {
    if (*active_it && (minimum == data_end || comparator(*data_it, *minimum)))
      minimum = data_it;
  }

  return minimum;
}

template<typename FORWARD_ITERATOR_DATA, typename FORWARD_ITERATOR_BOOL>
auto min_element(FORWARD_ITERATOR_DATA data_begin, FORWARD_ITERATOR_DATA data_end, FORWARD_ITERATOR_BOOL active_begin, FORWARD_ITERATOR_BOOL active_end)
{
  return best_element(data_begin, data_end, active_begin, active_begin, comparator_min<std::decay_t<decltype(*data_begin)>>());
}

template<typename FORWARD_ITERATOR_DATA, typename FORWARD_ITERATOR_BOOL>
auto max_element(FORWARD_ITERATOR_DATA data_begin, FORWARD_ITERATOR_DATA data_end, FORWARD_ITERATOR_BOOL active_begin, FORWARD_ITERATOR_BOOL active_end)
{
  return best_element(data_begin, data_end, active_begin, active_begin, comparator_max<std::decay_t<decltype(*data_begin)>>());
}

template<typename FORWARD_ITERATOR, typename COMPARATOR>
auto best_two_elements(FORWARD_ITERATOR begin, FORWARD_ITERATOR end, COMPARATOR comparator)
{
  auto first = end, second = end;

  for (auto it = begin; it != end; ++it) {
    if (first == end || comparator(*it, *first)) {
      second = first;
      first = it;
    } else if (second == end || comparator(*it, *second)) {
      second = it;
    }
  }

  return std::make_tuple(first, second);
}

template<typename FORWARD_ITERATOR>
auto least_two_elements(FORWARD_ITERATOR begin, FORWARD_ITERATOR end)
{
  return best_two_elements(begin, end, comparator_min<std::decay_t<decltype(*begin)>>());
}

template<typename FORWARD_ITERATOR>
auto greatest_two_elements(FORWARD_ITERATOR begin, FORWARD_ITERATOR end)
{
  return best_two_elements(begin, end, comparator_max<std::decay_t<decltype(*begin)>>());
}

template<typename FORWARD_ITERATOR, typename COMPARATOR>
auto best_two_values(FORWARD_ITERATOR begin, FORWARD_ITERATOR end, COMPARATOR comparator)
{
  auto [first, second] = best_two_elements(begin, end, comparator);

  const auto first_val = first != end ? *first : comparator.neutral();
  const auto second_val = second != end ? *second : comparator.neutral();

  return std::make_tuple(first_val, second_val);
}

template<typename FORWARD_ITERATOR>
auto least_two_values(FORWARD_ITERATOR begin, FORWARD_ITERATOR end)
{
  return best_two_values(begin, end, comparator_min<std::decay_t<decltype(*begin)>>());
}

template<typename FORWARD_ITERATOR>
auto greatest_two_values(FORWARD_ITERATOR begin, FORWARD_ITERATOR end)
{
  return best_two_values(begin, end, comparator_max<std::decay_t<decltype(*begin)>>());
}

}

#endif

/* vim: set ts=8 sts=2 sw=2 et ft=cpp: */
