#!/usr/bin/env python3

import argparse
import sys

from mpopt import utils, mwis

def parse_alpha(s):
    alpha = float(s)
    if alpha < 0.0 or alpha > 1.0:
        raise ValueError('Alpha must be between 0.0 and 1.0')
    return alpha


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='mwis_json', description='Optimizer for *.json maximum weighted independent set files.')
    parser.add_argument('-B', '--batch-size', type=int, default=mwis.DEFAULT_BATCH_SIZE)
    parser.add_argument('-b', '--max-batches', type=int, default=mwis.DEFAULT_MAX_BATCHES)
    parser.add_argument('-g', '--greedy-generations', type=int, default=mwis.DEFAULT_GREEDY_GENERATIONS, help='Specify number of greedy generation passes per batch.')
    parser.add_argument('-p', '--preprocess', action='store_true', help='Ignores nodes with positive cost.')
    parser.add_argument('-t', '--timeout', default=None)
    parser.add_argument('-a', '--alpha', type=parse_alpha)
    parser.add_argument('input_filename', metavar='INPUT', help='Specifies the *.jug input file.')
    args = parser.parse_args()

    with utils.smart_open(args.input_filename, 'rt') as f:
        model = mwis.deserialize(f)

    mapping = None
    if args.preprocess:
        model, mapping = mwis.persistency(model)

        original = len(mapping)
        reduced = sum(1 for x in mapping if x is None)
        print(f'Preprocessing: {original} -> {reduced} nodes ({reduced/original*100:.2f}%)')

    solver = mwis.construct_solver(model)

    if args.alpha:
        print("Setting alpha to", args.alpha)
        solver.alpha(args.alpha)

    if args.timeout:
        if args.timeout.endswith('h'):
            timeout = float(args.timeout[:-1]) * 3600
        elif args.timeout.endswith('m'):
            timeout = float(args.timeout[:-1]) * 60
        elif args.timeout.endswith('s'):
            timeout = float(args.timeout[:-1])
        else:
            timeout = float(args.timeout)

        import math
        import signal
        signal.alarm(math.ceil(timeout))

    del model
    import gc
    gc.collect()

    solver.run(args.batch_size, args.max_batches, args.greedy_generations)
