#!/usr/bin/env python3

import argparse
import json
import math
import signal

from mpopt import utils, mwis


def handle_multi_arg(params, default):
    set_params = [(param, value) for param, value in params if value is not None]

    if len(set_params) > 1:
        text = ', '.join([param for param, _ in set_params])
        raise RuntimeError(f"Specify only one of these arguments: {text}")

    if len(set_params) == 1:
        _, value = set_params[0]
    else:
        value = default

    print(f'  {params[0][0]} {value}')
    return value


def parse_time(s):
    if s.endswith('h'):
        return float(s[:-1]) * 3600
    elif s.endswith('m'):
        return float(s[:-1]) * 60
    elif s.endswith('s'):
        return float(s[:-1])
    else:
        return float(s)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='mwis_json', description='Optimizer for *.json maximum weighted independent set files.')
    parser.add_argument('-l', '--library', choices=('bregman_exp', 'bregman_log', 'temp_cont'), default='bregman_exp')
    parser.add_argument('-B', '--batch-size', type=int)
    parser.add_argument('-b', '--max-batches', type=int, default=100000)
    parser.add_argument('-g', '--greedy-generations', type=int, help='Specify number of greedy generation passes per batch.')
    parser.add_argument('-p', '--python', action='store_true', help='Use Python bregman implementation.')
    parser.add_argument('-t', '--timeout', type=parse_time, default=None)
    parser.add_argument('--initial-temperature', type=float)
    parser.add_argument('--threshold-feasibility', type=float)
    parser.add_argument('--threshold-stability', type=float)
    parser.add_argument('--temperature-drop-factor', type=float)
    parser.add_argument('--output-assignment', help='Writes the primal assignment to a file.')
    parser.add_argument('input_filename', metavar='INPUT', help='Specifies the *.jug input file.')
    # For paper: Additionally parse the arguments as referred to in the paper.
    parser.add_argument('--tau-batch', type=int)
    parser.add_argument('--tau-greedy', type=float)
    parser.add_argument('--tau-drop', type=float)
    parser.add_argument('--tau-feas', type=float)
    parser.add_argument('--tau-gap', type=float)
    parser.add_argument('--tau-stab', type=float)
    parser.add_argument('-T', type=float)
    args = parser.parse_args()

    # The helper `handle_multi_arg` also prints the parameters in the format
    # the paper uses. Therefore, we keep every invocation alphabetically sorted
    # by the first parameter (this one will get printed). Not all different
    # solver "libraries" accept all parameters though. We only handle and print
    # those that are actually used.
    print("Parameters:")
    print(f"  --library {args.library}")
    if args.library in ["temp_cont"]:
        args.batch_size = handle_multi_arg(
            [("--tau-batch", args.tau_batch), ("--batch-size", args.batch_size)],
            50)
        args.greedy_generations = handle_multi_arg(
            [("--tau-greedy", args.tau_batch), ("--greedy-generations", args.greedy_generations)],
            50)
    if args.library in ["bregman_log", "bregman_exp"]:
        args.greedy_generations = handle_multi_arg(
            [("--tau-batch", args.tau_batch), ("--batch-size", args.batch_size), ("--greedy-generations", args.greedy_generations)],
            50)
        args.batch_size = mwis.DEFAULT_BATCH_SIZE # We need something to pass to `solver.run(...)`.
    if args.library in ["bregman_log", "bregman_exp"]:
        args.temperature_drop_factor = handle_multi_arg(
            [("--tau-drop", args.tau_drop), ("--temperature-drop-factor", args.temperature_drop_factor)],
            0.5)
    if args.library in ["bregman_log", "bregman_exp"]:
        args.threshold_feasibility = handle_multi_arg(
            [("--tau-feas", args.tau_feas), ("--threshold-feasibility", args.threshold_feasibility)],
            0.01)
    if args.library in ["temp_cont"]:
        # NOTE: The C++ part actually uses the inverse values of the tau_gap parameter for historical
        #       reasons.
        args.temperature_drop_factor = handle_multi_arg(
            [("--tau-gap", 1 / args.tau_gap if args.tau_gap is not None else None), ("--temperature-drop-factor", args.temperature_drop_factor)],
            0.5)
    if args.library in ["bregman_exp"]:
        # TODO: For "temp_cont" no stabilisation is implemented. While it could be useful in practice,
        #       stabilization was not required for any benchmarked models with default parameters.
        args.threshold_stability = handle_multi_arg(
            [("--tau-stab", args.tau_stab), ("--threshold-stability", args.threshold_stability)],
            1e300)
    if args.library in ["bregman_log", "bregman_exp", "temp_cont"]:
        args.initial_temperature = handle_multi_arg(
            [("-T", args.T), ("--initial-temperature", args.initial_temperature)],
            0.01)
    print(flush=True)

    with utils.smart_open(args.input_filename, 'rt') as f:
        model = mwis.deserialize(f)

    solver = mwis.construct_solver(model, args.library)

    if args.initial_temperature:
        solver.temperature(args.initial_temperature)

    if args.threshold_feasibility:
        solver.threshold_feasibility(args.threshold_feasibility)

    if args.threshold_stability:
        solver.threshold_stability(args.threshold_stability)

    if args.temperature_drop_factor:
        solver.temperature_drop_factor(args.temperature_drop_factor)

    if args.timeout:
        signal.alarm(math.ceil(args.timeout))

    del model
    import gc
    gc.collect()

    try:
        solver.run(args.batch_size, args.max_batches, args.greedy_generations)
    except KeyboardInterrupt:
        # We ignore Ctrl-C here so that we store the primal assignment into a file even though the
        # user aborted the run with Ctrl-C.
        pass

    if args.output_assignment:
        print(f'Writing assignment to {args.output_assignment}...')
        with open(args.output_assignment, 'wt') as f:
            json.dump(assignment, f)
            f.write('\n')

    print('\nOk.')
